--  Copyright (C) 2003 David Roundy, 2010-2011 Petr Rockai
--
--  This program is free software; you can redistribute it and/or modify
--  it under the terms of the GNU General Public License as published by
--  the Free Software Foundation; either version 2, or (at your option)
--  any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU General Public License
--  along with this program; see the file COPYING.  If not, write to
--  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
--  Boston, MA 02110-1301, USA.
{-# LANGUAGE CPP, OverloadedStrings #-}

module Darcs.UI.Commands.Annotate ( annotate ) where

import Control.Arrow ( first )
import Control.Monad ( unless )

import Darcs.UI.Commands ( DarcsCommand(..), nodefaults, amInHashedRepository )
import Darcs.UI.Arguments
    ( DarcsFlag(..)
    , workingRepoDir
    , summary
    , unified
    , machineReadable
    , creatorhash
    , fixSubPaths
    , matchOne
    )
import Darcs.UI.Flags ( isUnified, toMatchFlags, useCache )
import Storage.Hashed.Plain( readPlainTree )
import Darcs.Repository.State ( readRecorded )
import Darcs.Repository
    ( Repository
    , withRepository
    , RepoJob(..)
    , readRepo
    , repoPatchType
    , listRegisteredFiles
    )
import Darcs.Patch.Set ( newset2RL )
import Darcs.Patch ( RepoPatch, Named, patch2patchinfo, invertRL )
import Darcs.Patch.Apply( ApplyState )
import qualified Darcs.Patch ( summary )
import Darcs.Patch.PatchInfoAnd ( PatchInfoAnd(..) )
import Darcs.Patch.Type ( PatchType(..) )
import Darcs.Patch.Dummy ( DummyPatch )
import qualified Data.ByteString.Char8 as BC ( pack, concat, intercalate )
import Data.ByteString.Lazy ( toChunks )
import Darcs.UI.PrintPatch ( printPatch, contextualPrintPatch )
import Darcs.Patch.ApplyMonad( withFileNames )
import System.FilePath.Posix ( (</>) )
import Darcs.Patch.Info ( showPatchInfoUI, showPatchInfo )
import Darcs.Patch.Match ( matchPatch, haveNonrangeMatch, getNonrangeMatchS  )
import Darcs.Repository.Match ( getFirstMatch, getOnePatchset )
import Darcs.Repository.Lock ( withTempDir )
import Darcs.Repository.PatchIndex ( getPatches, canUsePatchIndex )
import Darcs.Patch.Index.Types ( makePatchID )
import Darcs.Patch.Witnesses.Sealed ( Sealed2(..), Sealed(..), seal )
import Darcs.Patch.Witnesses.Ordered ( RL(..) )
import qualified Darcs.Patch.Annotate as A
import Darcs.Util.Printer ( putDocLn, Doc )

import Storage.Hashed.Tree( Tree, TreeItem(..), readBlob, list, expand )
import Storage.Hashed.Monad( findM, virtualTreeIO )
import Darcs.Util.Path( floatPath, anchorPath, fp2fn, toFilePath
                      , AbsolutePath )
import qualified Darcs.Util.Diff as D ( DiffAlgorithm(MyersDiff) )

#include "impossible.h"

annotateDescription :: String
annotateDescription = "Display which patch last modified something."

annotateHelp :: String
annotateHelp = unlines
 [ "The `darcs annotate` command provides two unrelated operations.  When"
 , "called on a file, it will find the patch that last modified each line"
 , "in that file.  When called on a patch (e.g. using `--patch`), it will"
 , "print the internal representation of that patch."
 , ""
 , "The `--summary` option will result in a summarized patch annotation,"
 , "similar to `darcs whatsnew`.  It has no effect on file annotations."
 , ""
 , "By default, output is in a human-readable format.  The `--machine-readable`"
 , "option can be used to generate output for machine postprocessing."
 , ""
 , "The `--creator-hash` option should only be used in combination with a"
 , "file or directory to be annotated.  In this case, the name of that file or"
 , "directory is interpreted to be its name *at the time it was created*"
 , "and the hash given along with `--creator-hash` indicates the patch"
 , "that created the file or directory.  This allows you to (relatively) easily"
 , "examine a file even if it has been renamed multiple times."
 ]

annotate :: DarcsCommand
annotate = DarcsCommand {commandProgramName = "darcs",
                         commandName = "annotate",
                         commandHelp = annotateHelp,
                         commandDescription = annotateDescription,
                         commandExtraArgs = -1,
                         commandExtraArgHelp = ["[FILE or DIRECTORY]..."],
                         commandCommand = annotateCmd,
                         commandPrereq = amInHashedRepository,
                         commandGetArgPossibilities = listRegisteredFiles,
                         commandArgdefaults = nodefaults,
                         commandAdvancedOptions = [],
                         commandBasicOptions = [summary,unified,
                                                 machineReadable,
                                                 matchOne, creatorhash,
                                                 workingRepoDir]}

annotateCmd :: (AbsolutePath, AbsolutePath) -> [DarcsFlag] -> [String] -> IO ()
annotateCmd fps opts files = withRepository (useCache opts) (RepoJob (annotate' fps opts files))

annotate' :: (RepoPatch p, ApplyState p ~ Tree)
          => (AbsolutePath, AbsolutePath)
          -> [DarcsFlag]
          -> [String]
          -> Repository p wR wU wR
          -> IO ()
annotate' _ opts [] repository = do
  let matchFlags = toMatchFlags opts
  unless (haveNonrangeMatch (PatchType :: PatchType DummyPatch) matchFlags) $
      fail $ "Annotate requires either a patch pattern or a " ++
               "file or directory argument."
  Sealed2 p <- matchPatch matchFlags `fmap` readRepo repository
  if Summary `elem` opts
     then do putDocLn $ showpi $ patch2patchinfo p
             putDocLn $ show_summary p
     else if isUnified opts
          then withTempDir "context" $ \_ ->
               do getFirstMatch repository matchFlags
                  c <- readPlainTree "."
                  contextualPrintPatch c p
          else printPatch p
    where showpi | MachineReadable `elem` opts = showPatchInfo
                 | otherwise                   = showPatchInfoUI
          show_summary :: RepoPatch p => Named p wX wY -> Doc
          show_summary = Darcs.Patch.summary

annotate' fps opts [""] repository = annotate' fps opts [] repository
annotate' fps opts args@[_] repository = do
  let matchFlags = toMatchFlags opts
  r <- readRepo repository
  (origpath:_) <- fixSubPaths fps args
  recorded <- readRecorded repository

  (patches, initial, path', patch_id) <-
    if haveNonrangeMatch (repoPatchType repository) matchFlags
       then do Sealed x <- getOnePatchset repository matchFlags
               let fn = [fp2fn $ toFilePath origpath]
                   nonRangeMatch = getNonrangeMatchS matchFlags r
                   (_, [path], _) = withFileNames Nothing fn nonRangeMatch
                   patch_id = Just $ (\(PIAP patchinfo _:<:_) -> makePatchID patchinfo) $ newset2RL x
               initial <- snd `fmap` virtualTreeIO (getNonrangeMatchS matchFlags r) recorded
               return (seal $ newset2RL x, initial, toFilePath path, patch_id)
       else return (seal $ newset2RL r, recorded, toFilePath origpath, Nothing)
  let path = "./" ++ path'

  found <- findM initial (floatPath $ toFilePath path)
  -- TODO need to decide about the --machine flag
  let fmt = if MachineReadable `elem` opts then A.machineFormat else A.format
  case found of
    Nothing -> fail $ "No such file or directory: " ++ toFilePath path
    Just (SubTree s) -> do
      s' <- expand s
      let subs = map (fp2fn . (path </>) . anchorPath "" . fst) $ list s'
          showPath (n, File _) = BC.pack (path </> n)
          showPath (n, _) = BC.concat [BC.pack (path </> n), "/"]
      (Sealed ans_patches) <- do
         upi <- canUsePatchIndex repository
         if not upi
            then return patches
            else getPatches repository subs patch_id
      putStrLn $ fmt (BC.intercalate "\n" $
                        map (showPath . first (anchorPath "")) $ list s') $
        A.annotateDirectory D.MyersDiff (invertRL ans_patches) (fp2fn path) subs
    Just (File b) -> do (Sealed ans_patches) <- do
                           upi <- canUsePatchIndex repository
                           if not upi
                              then return patches
                              else getPatches repository [fp2fn path] patch_id
                        con <- BC.concat `fmap` toChunks `fmap` readBlob b
                        putStrLn $ fmt con $ A.annotate D.MyersDiff (invertRL ans_patches) (fp2fn path) con
    Just (Stub _ _) -> impossible

annotate' _ _ _ _ = fail "annotate accepts at most one argument"
